from apscheduler.schedulers.background import BackgroundScheduler
from apscheduler.triggers.cron import CronTrigger
from apscheduler.triggers.interval import IntervalTrigger
from django_apscheduler.jobstores import DjangoJobStore
from apscheduler.executors.pool import ThreadPoolExecutor
from django.template import Template, Context
from clusterCore.api.serializer import object_to_json, c123_HostSerializer
from c123_kubernetes.tasks import drop_k8s_pg_pod, drop_mongo_k8s_pod
from .models import  c123_inventory, c123_node, c123_cluster, Task
from clusterCore.enums import  DBType, ClusterRoles, TaskState, Inventory, Status
from .tasks import  is_valid_job_target, find_replicaset_secondary, drop_mongo_node, drop_mariadb_node, drop_mysql_node, drop_elasticsearch_node, drop_redis_node, drop_pdb, drop_pg_node, check_if_pg_primary, drop_backup, send_mail, \
    update_mysql_fork_roles, update_mongo_roles, update_pg_roles, find_replicaset_primary, check_if_mysql_fork_primary, update_oracle_roles, update_elasticsearch_roles, remove_pg_slots
from django_apscheduler.models import DjangoJobExecution
from django.utils import timezone
from datetime import datetime
import json, smtplib, pytz
from django.db.models import F
from django.db import connection
from django_apscheduler.util import close_old_connections
import logging, atexit, sys
from django.contrib.auth.models import User

# general log, we need to know only if something are erroring (since "ok" informational messages bloating disk space much)
logging.basicConfig(filename='django_backend.log', filemode='a', format='%(asctime)s,%(msecs)d %(name)s %(levelname)s %(message)s', datefmt='%H:%M:%S', level=logging.ERROR)
scheduler_logger = logging.getLogger('apself.scheduler')
# logging of scheduler - here we need to know if scheduler are started plus the errors
scheduler_logger.setLevel(logging.INFO)
MISFIRE_GRACE_TIME_BACKUP = 36000
MISFIRE_GRACE_TIME_AUTO = 600
deletion_subject = "Reminder: Upcoming Deletion of {}"
deletion_notice = '''{{ subject }}
Dear {{ user }},
We hope this message finds you well. We are writing to remind you that in {{ days }} days, we will be performing routine maintenance that includes the deletion of some {{ object }} from our system.
As part of our regular maintenance process, we review and clean up inactive or outdated resources to ensure optimal performance and resource utilization. This action helps us maintain a secure and efficient environment for all our users.
If you have any concerns or specific requirements regarding the {{ object }} that might be affected, please let us know as soon as possible, and we will be more than happy to assist you.
Thank you for your cooperation. If you have any questions or need further assistance, don't hesitate to reach out to our support team.
Best regards,
{{ company }}
Support Team'''
@close_old_connections
def schedule_backup(task_id):
    task = Task.objects.filter(pk = task_id).first()
    if task:
        node = c123_node.objects.filter(pk = task.target_id, ind_deleted = False).first()
        #if we dont have a node to backup - no reason to continue
        if not node:
            return
        if not is_valid_job_target(node)[0]:
            return
        
        
        restore_tasks = [
            'pg_restore_task', 
            'pg_node_restore_task', 
            'pg_dumpall_restore_task', 
            'restore_elasticsearch_snapshot_task',
            'mysqldump_restore_task',
            'restore_mysql_node_task',
            'restore_mariadb_node_task',
            'import_sqlserver_instance_task',
            'import_sqlserver_database_task',
            'restore_sqlserver_instance_task',
            'mongo_backup_restore_task'
            'mongorestore_task',
            'restore_pdb_from_backup_task', 
            'restore_pdb_from_flashback_task',
            'import_pdb_task',
            'import_cdb_task'
        ]
        
        restore_in_progress = Task.objects.filter(target_id=node.pk,task__in=restore_tasks,status__in=[TaskState.PENDING.value, TaskState.RUNNING.value]).exists()
        
        if restore_in_progress:
            #better false-positive skip, than generate corrupted backups
            return
        
        match node.type:

            case DBType.POSTGRES_K8S.value:
                # task model using json string instead of json object, so need to convert to object...
                parameters = json.loads(task.parameters)
                old_backup = json.loads(parameters["backup_data"])
                backup = c123_inventory.objects.filter(pk=old_backup["id"]).first()
                if not backup:
                    return
                # old cluster.role
                backup_cluster_role = json.loads(parameters["node_data"])["cluster_role"]
                incremental_physical = backup.ex_params.get('incremental_physical')
                is_primary = check_if_pg_primary(node)
                # node got primary. but backup is for standby
                if is_primary is True and backup_cluster_role == ClusterRoles.STANDBY.value:
                    nodes = c123_node.objects.filter(cluster=node.cluster, type=DBType.POSTGRES_K8S.value,
                                                     ind_deleted=False).exclude(pk=node.pk)
                    # target backup to random standby
                    for cluster_node in nodes:
                        is_primary = check_if_pg_primary(cluster_node)
                        if is_primary is False:
                            now = timezone.now().strftime('%d%m%Y%H%M%S')  # time of new primary confirmed
                            backup.name = "{}_{}".format(cluster_node.name, now)
                            backup.node = cluster_node
                            backup.ex_params['backup_path'] = backup.ex_params['backup_path'].replace(
                                f'{node.host.name}_{node.port}', f'{cluster_node.host.name}_{cluster_node.port}')
                            if isinstance(incremental_physical, str):
                                backup.ex_params[
                                    "incremental_physical"] = f"{backup.ex_params['backup_path']}/{datetime.utcnow().strftime('%d%m%Y%H%M%S')}"
                            backup.save()
                            parameters["node_data"] = object_to_json(cluster_node)
                            parameters["host_data"] = c123_HostSerializer(cluster_node.host).data
                            parameters["backup_data"] = object_to_json(backup)
                            task.displayment = f"backup {cluster_node.cluster_role.lower()} {cluster_node.name}"
                            task.target_id = cluster_node.pk
                            task.parameters = json.dumps(parameters)
                            break
                # node got standby, but backup is for primary
                elif is_primary is False and backup_cluster_role == ClusterRoles.PRIMARY.value:
                    nodes = c123_node.objects.filter(cluster=node.cluster, type=DBType.POSTGRES_K8S.value,
                                                     ind_deleted=False).exclude(pk=node.pk)
                    # target backup to primary
                    for cluster_node in nodes:
                        is_primary = check_if_pg_primary(cluster_node)
                        if is_primary is True:
                            now = timezone.now().strftime('%d%m%Y%H%M%S')  # time of new primary confirmed
                            backup.name = "{}_{}".format(cluster_node.name, now)
                            backup.node = cluster_node
                            backup.ex_params['backup_path'] = backup.ex_params['backup_path'].replace(
                                f'{node.host.name}_{node.port}', f'{cluster_node.host.name}_{cluster_node.port}')
                            if isinstance(incremental_physical, str):
                                backup.ex_params[
                                    "incremental_physical"] = f"{backup.ex_params['backup_path']}/{datetime.utcnow().strftime('%d%m%Y%H%M%S')}"
                            backup.save()
                            parameters["node_data"] = object_to_json(cluster_node)
                            parameters["host_data"] = c123_HostSerializer(cluster_node.host).data
                            parameters["backup_data"] = object_to_json(backup)
                            task.displayment = f"backup {cluster_node.cluster_role.lower()} {cluster_node.name}"
                            task.target_id = cluster_node.pk
                            task.parameters = json.dumps(parameters)
                            break
                else:
                    now = timezone.now().strftime('%d%m%Y%H%M%S')
                    backup.name = "{}_{}".format(node.name, now)
                    backup.save()
                    parameters["node_data"] = object_to_json(node)
                    parameters["backup_data"] = object_to_json(backup)
                    task.parameters = json.dumps(parameters)


            case DBType.POSTGRES.value:
                #task model using json string instead of json object, so need to convert to object...
                parameters = json.loads(task.parameters)
                old_backup = json.loads(parameters["backup_data"])
                backup = c123_inventory.objects.filter(pk = old_backup["id"]).first()
                if not backup:
                    return
                #old cluster.role
                backup_cluster_role = json.loads(parameters["node_data"])["cluster_role"]
                incremental_physical = backup.ex_params.get('incremental_physical')
                is_primary = check_if_pg_primary(node)
                #node got primary. but backup is for standby
                if is_primary is True and backup_cluster_role == ClusterRoles.STANDBY.value:
                    nodes = c123_node.objects.filter(cluster = node.cluster, type = DBType.POSTGRES.value, ind_deleted = False).exclude(pk = node.pk)
                    #target backup to random standby
                    for cluster_node in nodes:
                           is_primary = check_if_pg_primary(cluster_node)
                           if is_primary is False:
                               now = timezone.now().strftime('%d%m%Y%H%M%S') #time of new primary confirmed
                               backup.name = "{}_{}".format(cluster_node.name, now)
                               backup.node = cluster_node
                               backup.ex_params['backup_path'] = backup.ex_params['backup_path'].replace(f'{node.host.name}_{node.port}', f'{cluster_node.host.name}_{cluster_node.port}') 
                               if isinstance(incremental_physical, str):
                                    backup.ex_params["incremental_physical"] = f"{backup.ex_params['backup_path']}/{datetime.utcnow().strftime('%d%m%Y%H%M%S')}" 
                               backup.save()
                               parameters["node_data"] = object_to_json(cluster_node)
                               parameters["host_data"] = c123_HostSerializer(cluster_node.host).data
                               parameters["backup_data"] = object_to_json(backup)
                               task.displayment = f"backup {cluster_node.cluster_role.lower()} {cluster_node.name}"
                               task.target_id = cluster_node.pk
                               task.parameters = json.dumps(parameters)
                               break
                #node got standby, but backup is for primary
                elif is_primary is False and backup_cluster_role == ClusterRoles.PRIMARY.value:
                     nodes = c123_node.objects.filter(cluster = node.cluster, type = DBType.POSTGRES.value, ind_deleted = False).exclude(pk = node.pk)
                     #target backup to primary
                     for cluster_node in nodes:
                           is_primary = check_if_pg_primary(cluster_node)
                           if is_primary is True:
                               now = timezone.now().strftime('%d%m%Y%H%M%S') #time of new primary confirmed
                               backup.name = "{}_{}".format(cluster_node.name, now)
                               backup.node = cluster_node
                               backup.ex_params['backup_path'] = backup.ex_params['backup_path'].replace(f'{node.host.name}_{node.port}', f'{cluster_node.host.name}_{cluster_node.port}') 
                               if isinstance(incremental_physical, str):
                                    backup.ex_params["incremental_physical"] = f"{backup.ex_params['backup_path']}/{datetime.utcnow().strftime('%d%m%Y%H%M%S')}" 
                               backup.save()
                               parameters["node_data"] = object_to_json(cluster_node)
                               parameters["host_data"] = c123_HostSerializer(cluster_node.host).data
                               parameters["backup_data"] = object_to_json(backup)
                               task.displayment = f"backup {cluster_node.cluster_role.lower()} {cluster_node.name}"
                               task.target_id = cluster_node.pk
                               task.parameters = json.dumps(parameters)
                               break
                else:
                    now = timezone.now().strftime('%d%m%Y%H%M%S')
                    backup.name = "{}_{}".format(node.name, now)
                    backup.save()
                    parameters["node_data"] = object_to_json(node)
                    parameters["backup_data"] = object_to_json(backup)
                    task.parameters = json.dumps(parameters) 

            case DBType.MONGO.value:
                parameters = json.loads(task.parameters)
                old_backup = json.loads(parameters["backup_data"])
                backup = c123_inventory.objects.filter(pk = old_backup["id"]).first()
                if not backup:
                    return
                
                backup_cluster_role = json.loads(parameters["node_data"])["cluster_role"]
                nodes = c123_node.objects.filter(cluster = node.cluster, type = DBType.MONGO.value, ind_deleted = False).exclude(cluster_role = ClusterRoles.ARBITER.value)
                #backup are for primary - target backup to primary
                if backup_cluster_role == ClusterRoles.PRIMARY.value:
                    for cluster_node in nodes:
                        primary_name = find_replicaset_primary(cluster_node)
                        if primary_name != 'none':
                            break
                    #nothing changed, update just backup date
                    if primary_name == node.name:
                        now = timezone.now().strftime('%d%m%Y%H%M%S')
                        backup.name = "{}_{}".format(node.name, now)
                        backup.save()
                        parameters["node_data"] = object_to_json(node)
                        parameters["backup_data"] = object_to_json(backup)
                        task.parameters = json.dumps(parameters) 
                    else:
                        primary = c123_node.objects.filter(type = DBType.MONGO.value, cluster_role = ClusterRoles.PRIMARY.value, ind_deleted = False, name = primary_name).first()
                        if primary:
                            standbys = nodes.filter(cluster_role = ClusterRoles.STANDBY.value)
                            arbiters = nodes.filter(cluster_role = ClusterRoles.ARBITER.value)
                            arbiters_id = [item.pk for item in arbiters]
                            secondary_data = [{"id": item.pk, "name": item.name } for item in standbys]
                            parameters['secondary_data'] = secondary_data
                            parameters['arbiters_id'] = arbiters_id
                            now = timezone.now().strftime('%d%m%Y%H%M%S') #time of new primary confirmed
                            backup.name = "{}_{}".format(primary.name, now)
                            backup.node = primary
                            backup.ex_params['backup_path'] = backup.ex_params['backup_path'].replace(f'{node.host.name}_{node.port}', f'{primary.host.name}_{primary.port}') 
                            backup.save()
                            parameters["node_data"] = object_to_json(primary)
                            parameters["host_data"] = c123_HostSerializer(primary.host).data
                            parameters["backup_data"] = object_to_json(backup)
                            task.displayment = f"backup {primary.cluster_role.lower()} {primary.name}"
                            task.target_id = primary.pk
                            task.parameters = json.dumps(parameters)
                
                #backup are for secondary - target backup to random secondary
                else:
                    secondary_name = 'none'
                    for cluster_node in nodes:
                        secondary_name = find_replicaset_secondary(cluster_node, node.name)
                        if secondary_name != 'none':
                            break
                    target_secondary = nodes.filter(name = secondary_name).first()
                    if target_secondary:
                        now = timezone.now().strftime('%d%m%Y%H%M%S') #time of secondary confirmed
                        backup.name = "{}_{}".format(secondary_name, now)
                        backup.node = target_secondary
                        backup.ex_params['backup_path'] = backup.ex_params['backup_path'].replace(f'{node.host.name}_{node.port}', f'{target_secondary.host.name}_{target_secondary.port}') 
                        backup.save()
                        parameters["node_data"] = object_to_json(target_secondary)
                        parameters["host_data"] = c123_HostSerializer(target_secondary.host).data
                        parameters["backup_data"] = object_to_json(backup)
                        task.displayment = f"backup {target_secondary.cluster_role.lower()} {target_secondary.name}"
                        task.target_id = target_secondary.pk
                        task.parameters = json.dumps(parameters)
            case DBType.MYSQL.value | DBType.MARIADB.value:
                parameters = json.loads(task.parameters)
                old_backup = json.loads(parameters["backup_data"])
                backup = c123_inventory.objects.filter(pk = old_backup["id"]).first()
                incremental = backup.ex_params.get("incremental")
                if incremental:
                    parameters['backup_data'] = object_to_json(backup)
                    task.parameters = json.dumps(parameters)
                    from django.conf import settings
                    job = settings.scheduler.get_job(job_id=backup.ex_params.get("job_id"))
                    if job:
                        if isinstance(job.trigger, CronTrigger):
                            backup_interval = backup.ex_params.get("incremental_interval")
                            job.reschedule(trigger = IntervalTrigger(minutes=(int(backup_interval)) if backup_interval else 30))
                backup_cluster_role = json.loads(parameters["node_data"])["cluster_role"]
                is_primary = check_if_mysql_fork_primary(node)
                #node got primary. but backup is for standby
                if is_primary is True and backup_cluster_role == ClusterRoles.STANDBY.value:
                    nodes = c123_node.objects.filter(cluster = node.cluster, type = node.type, ind_deleted = False).exclude(pk = node.pk)
                    #target backup to random standby
                    for cluster_node in nodes:
                           is_primary = check_if_mysql_fork_primary(cluster_node)
                           if is_primary is False:
                               now = timezone.now().strftime('%d%m%Y%H%M%S') #time of new primary confirmed
                               backup.name = "{}_{}".format(cluster_node.name, now)
                               backup.node = cluster_node
                               backup.ex_params['backup_path'] = backup.ex_params['backup_path'].replace(f'{node.host.name}_{node.port}', f'{cluster_node.host.name}_{cluster_node.port}') 
                               if isinstance(incremental, str):
                                   incremental = incremental.replace(f'{node.host.name}_{node.port}', f'{cluster_node.host.name}_{cluster_node.port}')
                                   backup.ex_params["incremental"] = incremental 
                               backup.save()
                               parameters["node_data"] = object_to_json(cluster_node)
                               parameters["host_data"] = c123_HostSerializer(cluster_node.host).data
                               parameters["backup_data"] = object_to_json(backup)
                               task.displayment = f"backup {cluster_node.cluster_role.lower()} {cluster_node.name}"
                               task.target_id = cluster_node.pk
                               task.parameters = json.dumps(parameters)
                               break
                #node got standby, but backup is for primary
                elif is_primary is False and backup_cluster_role == ClusterRoles.PRIMARY.value:
                     nodes = c123_node.objects.filter(cluster = node.cluster, type = node.type, ind_deleted = False).exclude(pk = node.pk)
                     #target backup to primary
                     for cluster_node in nodes:
                           is_primary = check_if_mysql_fork_primary(cluster_node)
                           if is_primary is True:
                               now = timezone.now().strftime('%d%m%Y%H%M%S') #time of new primary confirmed
                               backup.name = "{}_{}".format(cluster_node.name, now)
                               backup.node = cluster_node
                               backup.ex_params['backup_path'] = backup.ex_params['backup_path'].replace(f'{node.host.name}_{node.port}', f'{cluster_node.host.name}_{cluster_node.port}') 
                               if isinstance(incremental, str):
                                   incremental = incremental.replace(f'{node.host.name}_{node.port}', f'{cluster_node.host.name}_{cluster_node.port}')
                                   backup.ex_params["incremental"] = incremental 
                               backup.save()
                               parameters["node_data"] = object_to_json(cluster_node)
                               parameters["host_data"] = c123_HostSerializer(cluster_node.host).data
                               parameters["backup_data"] = object_to_json(backup)
                               task.displayment = f"backup {cluster_node.cluster_role.lower()} {cluster_node.name}"
                               task.target_id = cluster_node.pk
                               task.parameters = json.dumps(parameters)
                               break
                else:
                    now = timezone.now().strftime('%d%m%Y%H%M%S')
                    backup.name = "{}_{}".format(node.name, now)
                    backup.save()
                    parameters["node_data"] = object_to_json(node)
                    parameters["backup_data"] = object_to_json(backup)
                    task.parameters = json.dumps(parameters) 
        if task.status != TaskState.RUNNING.value:
            task.status = TaskState.PENDING.value
        
        task.save()

@close_old_connections
def mongodump_oplog_job(cluster_id):
    nodes = c123_node.objects.filter(cluster__id = cluster_id, type = DBType.MONGO.value, ind_deleted = False)
    task = None
    for node in nodes:
        task = Task.objects.filter(task = "mongodump_oplog_task", target_id = node.id)
        if task.exists():
            break
    else:
        from django.conf import settings
        jobs = settings.scheduler.get_jobs()
        for job in jobs:
            if f"mongodump_oplog_task_{cluster_id}" in job.id:
                settings.scheduler.remove_job(job.id)
        return
    
    task = task.first()
    if task:
            node = nodes.filter(pk = task.target_id, ind_deleted = False).first() 
            if node:
                if not is_valid_job_target(node)[0]:
                    return
            if task.status != TaskState.RUNNING.value:
                    if Task.objects.filter(target_id = task.target_id, task = 'physical_backup_mongo_node_task', status__in=[TaskState.PENDING.value, TaskState.RUNNING.value]).exists() or Task.objects.filter(task="mongo_backup_restore_task", status__in=[TaskState.PENDING.value, TaskState.RUNNING.value], target_id = task.target_id).exists():
                        return
                    else:
                          primary = c123_node.objects.filter(cluster = node.cluster, cluster_role = ClusterRoles.PRIMARY.value, ind_deleted = False).first() if node.cluster_role != ClusterRoles.PRIMARY.value else node
                          master = find_replicaset_primary(node)
                          if primary.name in master:
                              parameters = json.loads(task.parameters)
                              parameters["primary_data"] = object_to_json(primary)
                              parameters["host_primary"] = c123_HostSerializer(primary.host).data
                              old_backup = json.loads(parameters["backup_data"])
                              backup = c123_inventory.objects.filter(pk = old_backup["id"]).first()
                              if backup:
                                  old_backup["ex_params"]['backup_path'] = backup.ex_params.get('backup_path')
                                  parameters["backup_data"] = json.dumps(old_backup)
                              task.parameters = json.dumps(parameters)
                              task.target_id = primary.pk
                              task.save()
                          elif primary.name not in master:
                              for cluster_node in nodes.exclude(pk = node.id):
                                  master = find_replicaset_primary(cluster_node)
                                  primary = c123_node.objects.filter(name = master).first()
                                  if primary and primary.cluster_role != ClusterRoles.PRIMARY.value:
                                      primary.cluster_role = ClusterRoles.PRIMARY.value
                                      primary.ex_params["mongo_master_id"] = primary.id
                                      primary.save()
                                      others = nodes.exclude(pk = primary.pk)
                                      for other in others:
                                          if other.cluster_role == ClusterRoles.PRIMARY.value:
                                              other.cluster_role = ClusterRoles.STANDBY.value
                                          other.ex_params["mongo_master_id"] = primary.id
                                          other.save()
                                      parameters = json.loads(task.parameters)
                                      parameters["primary_data"] = object_to_json(primary)
                                      parameters["host_primary"] = c123_HostSerializer(primary.host).data
                                      old_backup = json.loads(parameters["backup_data"])
                                      backup = c123_inventory.objects.filter(pk = old_backup["id"]).first()
                                      if backup:
                                          old_backup["ex_params"]['backup_path'] = backup.ex_params.get('backup_path')
                                          parameters["backup_data"] = json.dumps(old_backup)
                                      task.parameters = json.dumps(parameters)
                                      task.target_id = primary.pk
                                      task.save()
                                      break        
@close_old_connections       
def UpdateMySQLForkClusterRoles(cluster_id):
    app_primary = c123_node.objects.filter(type__in=[DBType.MARIADB.value, DBType.MYSQL.value], cluster_role = ClusterRoles.PRIMARY.value, cluster__id = cluster_id, ind_deleted = False).exclude(status = Status.DROPPING.value).first()
    if not app_primary:
        return
    update_mysql_fork_roles(app_primary)


@close_old_connections       
def UpdateOracleClusterRoles(cluster_id):
    cdb = c123_node.objects.filter(type = DBType.ORACLE_CDB.value, cluster__id = cluster_id, ind_deleted = False).first()
    if not cdb:
        return
    update_oracle_roles(cdb)

@close_old_connections       
def UpdatePgClusterRoles(cluster_id):
    app_primary = c123_node.objects.filter(type = DBType.POSTGRES.value, cluster_role = ClusterRoles.PRIMARY.value, cluster__id = cluster_id, ind_deleted = False).exclude(status = Status.DROPPING.value).first()
    if not app_primary:
        return
    update_pg_roles(app_primary)


@close_old_connections       
def RemovePgSlots(cluster_id):
    cluster = c123_cluster.objects.filter(type = DBType.POSTGRES.value, id = cluster_id).first()
    if not cluster:
        return
    remove_pg_slots(cluster)

@close_old_connections       
def UpdateElasticsearchClusterRoles(cluster_id):
    primary = c123_node.objects.filter(type = DBType.ELASTIC.value, cluster_role = ClusterRoles.PRIMARY.value, cluster__id = cluster_id, ind_deleted = False).exclude(status = Status.DROPPING.value).first()
    if not primary:
        return
    
    update_elasticsearch_roles(primary)

@close_old_connections       
def UpdateMongoClusterRoles(cluster_id):
    app_primary = c123_node.objects.filter(type = DBType.MONGO.value, cluster_role = ClusterRoles.PRIMARY.value, cluster__id = cluster_id, ind_deleted = False).exclude(status = Status.DROPPING.value).first()
    if not app_primary:
        return
    update_mongo_roles(app_primary)

class SchedulerWrapper(object):
    
    def __init__(self) -> None:
        try:
            terminal_arguments = sys.argv
            if any(a in terminal_arguments for a in ['makemigrations', 'migrate']):
                return
            job_defaults = {
                'coalesce': True,
                'misfire_grace_time': MISFIRE_GRACE_TIME_AUTO,
                'max_instances': 1
            }
            from django.conf import settings
            self.scheduler = BackgroundScheduler(timezone = settings.TIME_ZONE, job_defaults = job_defaults)
            self.scheduler.add_jobstore(DjangoJobStore(), "default")
            self.scheduler.add_executor(ThreadPoolExecutor(10), "default")
            self.scheduler.add_job(auto_delete_nodes, id ='auto_delete_nodes', trigger=CronTrigger(hour="09", minute="00"), replace_existing=True, misfire_grace_time=MISFIRE_GRACE_TIME_AUTO)
            self.scheduler.add_job(delete_old_inventory, id = 'delete_old_inventory', trigger= CronTrigger(hour="09", minute="00") , replace_existing = True, misfire_grace_time=MISFIRE_GRACE_TIME_AUTO)
            self.scheduler.add_job(rp_mail, id='rp_mail', trigger=IntervalTrigger(minutes=10),replace_existing = True, misfire_grace_time=MISFIRE_GRACE_TIME_AUTO)
            self.scheduler.add_job(delete_old_job_executions, id = 'delete_old_job_executions', trigger= CronTrigger(day_of_week="mon", hour="00", minute="00"),replace_existing = True, misfire_grace_time=MISFIRE_GRACE_TIME_AUTO)
            if not self.scheduler.running:
                self.scheduler.start()
            atexit.register(self.shutdown)
            scheduler_logger.info("Job scheduler initialization ok")
        except Exception as e:
            scheduler_logger.error(f"Scheduler: {str(e)}")
            if self.scheduler.running:
                self.scheduler.shutdown(wait = True)
            if not self.scheduler.running:
                self.scheduler.start()
    
    def add_job(self, func, *args, **kwargs):
        try:
            self.scheduler.add_job(func, *args, **kwargs)
        except Exception as e:
             scheduler_logger.error(f"Scheduler: {str(e)}")

    
    def get_job(self, job_id):
        try:
            return self.scheduler.get_job(job_id)
        except Exception as e:
            scheduler_logger.error(f"Scheduler: {str(e)}")

    def get_jobs(self):
        try:
            return self.scheduler.get_jobs()
        except Exception as e:
            scheduler_logger.error(f"Scheduler: {str(e)}")

    def remove_job(self, job_id):
        try:
            self.scheduler.remove_job(job_id)
        except Exception as e:
            scheduler_logger.error(f"Scheduler: {str(e)}")
    
    def shutdown(self):
        try:
            if self.scheduler.running:
                self.scheduler.shutdown()
                scheduler_logger.info("Job scheduler shutdown gracefully")
        except Exception as e:
            scheduler_logger.error(f"Scheduler: {str(e)}")

@close_old_connections
def auto_delete_nodes():
    now = timezone.now()
    notify_targets = c123_node.objects.filter(ind_deleted = False, auto_delete__isnull = False, auto_delete__lte = (now + timezone.timedelta(days = 14)).date())
    for target in notify_targets:
        try:
            from django.conf import settings
            days = (datetime.combine(target.auto_delete, now.time()).astimezone(pytz.timezone(settings.TIME_ZONE)) - now).days
            if days in [13, 6, 0]:
                user = target.created_by
                username = user.username
                email = user.email
                subject = deletion_subject.format("nodes")
                template = Template(deletion_notice)
                context = Context({"subject": subject, 'object': 'nodes', "user": username, "days": days + 1, "company": "123Cluster"})
                message = template.render(context)
                send_mail(subject, message, [email])
        except Exception as e:
            print('auto_delete_nodes:', str(e))
    
    nodes = c123_node.objects.filter(ind_deleted = False, auto_delete__isnull = False, auto_delete__lte = now.date())
    if not nodes.exists():
        return
    
    for node in nodes:
        match node.type:
            case DBType.REDIS.value:
                task_exists = Task.objects.filter(target_id = node.id, task = "drop_redis_node_task", status__in=[TaskState.PENDING.value, TaskState.RUNNING.value, TaskState.ON_DEMAND.value]).exists()
                if not task_exists:
                    drop_redis_node(node)
            
            case DBType.MYSQL.value:
                task_exists = Task.objects.filter(target_id = node.id, task = "drop_mysql_node_task", status__in=[TaskState.PENDING.value, TaskState.RUNNING.value, TaskState.ON_DEMAND.value]).exists()
                if not task_exists:
                    drop_mysql_node(node)
            
            case DBType.MONGO.value:
                task_exists = Task.objects.filter(target_id = node.id, task__in = ["delete_mongo_standalone_node_task", "delete_mongo_replicaset_node_tesk"], status__in=[TaskState.PENDING.value, TaskState.RUNNING.value, TaskState.ON_DEMAND.value]).exists()
                if not task_exists:
                    drop_mongo_node(node)

            case DBType.MONGO_K8S.value:
                task_exists = Task.objects.filter(target_id=node.id, task__in=["uninstall_mongodb_k8s_standalone_task"],
                                                  status__in=[TaskState.PENDING.value, TaskState.RUNNING.value,
                                                              TaskState.ON_DEMAND.value]).exists()
                if not task_exists:
                    drop_mongo_k8s_pod(node)
            
            case DBType.MARIADB.value:
                task_exists = Task.objects.filter(target_id = node.id, task = "drop_mariadb_node_task", status__in=[TaskState.PENDING.value, TaskState.RUNNING.value, TaskState.ON_DEMAND.value]).exists()
                if not task_exists:
                    drop_mariadb_node(node)

            case DBType.ELASTIC.value:
                task_exists = Task.objects.filter(target_id = node.id, task = "drop_elasticsearch_node_task", status__in=[TaskState.PENDING.value, TaskState.RUNNING.value, TaskState.ON_DEMAND.value]).exists()
                if not task_exists:
                    drop_elasticsearch_node(node)
            
            case DBType.POSTGRES.value:
                task_exists = Task.objects.filter(target_id = node.id, task = "drop_pg_node_task", status__in=[TaskState.PENDING.value, TaskState.RUNNING.value, TaskState.ON_DEMAND.value]).exists()
                if not task_exists:
                    drop_pg_node(node)

            case DBType.POSTGRES_K8S.value:
                task_exists = Task.objects.filter(target_id = node.id, task = "drop_k8s_pg_pod_task", status__in=[TaskState.PENDING.value, TaskState.RUNNING.value, TaskState.ON_DEMAND.value]).exists()
                if not task_exists:
                    drop_k8s_pg_pod(node)
            
            case DBType.ORACLE_PDB.value:
                task_exists = Task.objects.filter(target_id = node.id, task = "drop_pdb_task", status__in=[TaskState.PENDING.value, TaskState.RUNNING.value, TaskState.ON_DEMAND.value]).exists()
                if not task_exists:
                    cdb = c123_node.objects.filter(pk = node.ex_params.get("cdb_id")).first()
                    if not cdb:
                        print("auto_delete_nodes:", f"{node.name} does not belong to any cdb")
                    else:
                        drop_pdb(cdb, node)                            
    
    User.objects.filter(
    id__in=nodes.values_list("created_by", flat=True).distinct(),
    profile__is_demo=True).delete()

@close_old_connections
def delete_old_job_executions(max_age=604_800):
    DjangoJobExecution.objects.delete_old_job_executions(max_age)

@close_old_connections
def rp_mail():
    try:
        cursor = connection.cursor()
        cursor.execute("SELECT * FROM app_123monitor.rp_mail_queue WHERE ind_sent = 'P' ORDER BY date_created DESC;")
    except Exception as error:
        logging.error(f"rp_mail: {str(error)}")
        return
    
    try:
        queue = cursor.fetchall()
    except Exception as error:
        logging.error(f"rp_mail: {str(error)}")
        return
    
    for i, item in enumerate(queue):
        if i == 10:
            break
        
        try:
            status = send_mail(subject=item[5], body=item[6], to=[item[4]])
        except smtplib.SMTPDataError as e:
            print(e)
            try:
                cursor.execute(f"UPDATE app_123monitor.rp_mail_queue SET ind_sent = 'F' WHERE id ={item[0]}")
            except Exception as error:
                logging.error(f"rp_mail: {str(error)}")
                return
        
        if status != 250:
            try:
                cursor.execute(f"UPDATE app_123monitor.rp_mail_queue SET ind_sent = 'N' WHERE id ={item[0]}")
            except Exception as error:
                logging.error(f"rp_mail: {str(error)}")
                return
        else:
            try:
                cursor.execute("UPDATE app_123monitor.rp_mail_queue SET ind_sent = 'S', date_sent = %s WHERE id = %s",[item[0], timezone.now()])
            except Exception as error:
                logging.error(f"rp_mail: {str(error)}")
                return

@close_old_connections
def delete_old_inventory():
    now = timezone.now()
    notify_targets = c123_inventory.objects.filter(type__in=[Inventory.BACKUP.value, Inventory.EXPORT.value], created_date__lte=now - timezone.timedelta(days=1) * F("keep_days") + timezone.timedelta(days=14))
    for target in notify_targets:
        days = ((target.created_date + timezone.timedelta(days = target.keep_days)) - now).days
        if days in [13, 6, 0]:
            try:
                user = target.created_by
                username = user.username
                email = user.email
                subject = deletion_subject.format("backups")
                template = Template(deletion_notice)
                context = Context({"subject": subject, 'object': 'backups', "user": username, "days": days + 1, "company": "123Cluster"})
                message = template.render(context)
                send_mail(subject, message, [email])
                print('mail has sent')
            except Exception as e:
                print("delete_old_inventory:", str(e))
    
    inventory = c123_inventory.objects.filter(type__in=[Inventory.BACKUP.value, Inventory.EXPORT.value], created_date__lte=now - timezone.timedelta(days=1) * F("keep_days"))
    drop_backup(inventory, automate=True)   